/*=====================================================================
  File:      FindType.cool

  Summary:   Command Line utility to help find and display information
             on COM+ 2.0 types.

  Classes:   FindType
			 Indented Writer

  Origin:    COM+ 2.0 Samples Team

---------------------------------------------------------------------
  This file is part of the Microsoft COM+ 2.0 SDK Code Samples.

  Copyright (C) 1999 Microsoft Corporation.  All rights reserved.

This source code is intended only as a supplement to Microsoft
Development Tools and/or on-line documentation.  See these other
materials for detailed information regarding Microsoft code samples.

THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
PARTICULAR PURPOSE.
=====================================================================*/
namespace Microsoft.Samples.Reflection
{	
	using System;
	using System.Reflection;
	using System.IO;
	using System.Collections;
	using System.Text;
	using System.Runtime.InteropServices;
	/*=====================================================================
	  Class:    FindType 

	  Summary:  Utility class that holds methods that can be used to find and 
				display information about types. Use this utility to:
				
				(1) Find a specific type
				(2) Optionally display information about the type including 
				    interfaces, properties, events and methods
				
				This utility dumps information directly contained in the type 
				specified. In order to get a complete dump of a type and all 
				base type information you must use the "recurse" mode of the 
				utility.
				
	  Note:     This class searches for types based on knowledge of how the
	            current class loader works. As such it is highly likely that
	            this utility will not work correctly if the class loader logic 
	            fails. 
	            
	            In particular the current version of this utility does not 
	            search the Fusion cache.

	=====================================================================*/
	public class FindType
	{
		private const int SHOW_INTERFACES  = 0x01;		
		private const int SHOW_FIELDS      = 0x02;
		private const int SHOW_PROPERTIES  = 0x04;
		private const int SHOW_EVENTS      = 0x08;
		private const int SHOW_METHODS     = 0x10;
		private const int SHOW_MODULE_INFO = 0x20;
	
		private IndentedWriter myVerboseWriter = new IndentedWriter();
		private IndentedWriter myWriter        = new IndentedWriter();
		
		private bool exactMatchOnly      = false;
		private bool recurseTypes        = false;
		private bool matchOnlyNamespace  = false;
		private bool wideSearch          = false;
				
		private int showOptions = 0;
				
		private ArrayList myClassList    = new ArrayList();
		private ObjectList DirList		 = new ObjectList();

		/*=====================================================================
		  [Public] Method: FindType (constructor)

		  Summary:  Constructor for class.

		  Params:   None.
		=====================================================================*/
	    public FindType() 
		{
	    }    
	    
		/*=====================================================================
		  [Public] Property: ExactMatchOnly

		  Type:	Boolean  

		  Description:   Do an exact match of passed types - the name passed must 
		                 be the same as the fully qualified type name.
		=====================================================================*/
		public bool ExactMatchOnly
		{
			get { return exactMatchOnly; }
			set { exactMatchOnly = value; }
		}


		/*=====================================================================
		  [Public] Property: RecurseTypes

		  Type:	Boolean  

		  Description:   Load all base types and display the same information 
		                 (methods, properties etc) for the base type.
		=====================================================================*/
		public bool RecurseTypes
		{
			get { return recurseTypes; }
			set { recurseTypes = value; }
		}
						
		/*=====================================================================
		  [Public] Property: MatchOnlyNamespace

		  Type:	Boolean  

		  Description:   Match only against the namespace component of a type.
						 Use this to display all members of a namespace.
		=====================================================================*/
		public bool MatchOnlyNamespace
		{
			get { return matchOnlyNamespace; }
			set { matchOnlyNamespace = value; }
		}

		/*=====================================================================
		  [Public] Property: WideSearch 

		  Type:	Boolean  

		  Description:   Match anywhere in the name (namespace and type name).
		=====================================================================*/
		public bool WideSearch
		{
			get { return wideSearch; }
			set { wideSearch = value; }
		}
		
		/*=====================================================================
		  [Public] Property: Show Interfaces

		  Type:	Boolean  

		  Description:   Show any interfaces implemented by this type.
		=====================================================================*/
		public bool ShowInterfaces
		{
			get { return (showOptions & SHOW_INTERFACES) != 0; }
			set { 
					if ( value )
						showOptions |= SHOW_INTERFACES; 
					else
						showOptions &= ~SHOW_INTERFACES; 
				}
		}
		
		/*=====================================================================
		  [Public] Property: Show Fields

		  Type:	Boolean  

		  Description:   Show any public fields in the type found.
		=====================================================================*/
		public bool ShowFields
		{
			get { return (showOptions & SHOW_FIELDS) != 0; }
			set { 
					if ( value )
						showOptions |= SHOW_FIELDS; 
					else
						showOptions &= ~SHOW_FIELDS; 
				}
		}
		
		/*=====================================================================
		  [Public] Property: Show Properties

		  Type:	Boolean  

		  Description:   Show any public properties in the type found.
		=====================================================================*/
		public bool ShowProperties
		{
			get { return (showOptions & SHOW_PROPERTIES) != 0; }
			set { 
					if ( value )
						showOptions |= SHOW_PROPERTIES; 
					else
						showOptions &= ~SHOW_PROPERTIES; 
				}
		}
		
		/*=====================================================================
		  [Public] Property: Show Events

		  Type:	Boolean  

		  Description:   Show any public events in the type found.
		=====================================================================*/
		public bool ShowEvents
		{
			get { return (showOptions & SHOW_EVENTS) != 0; }
			set { 
					if ( value )
						showOptions |= SHOW_EVENTS; 
					else
						showOptions &= ~SHOW_EVENTS; 
				}
		}
		
		/*=====================================================================
		  [Public] Property: Show Methods

		  Type:	Boolean  

		  Description:   Show any public methods in the type found.
		=====================================================================*/
		public bool ShowMethods
		{
			get { return (showOptions & SHOW_METHODS) != 0 ; }
			set { 
					if ( value )
						showOptions |= SHOW_METHODS; 
					else
						showOptions &= ~SHOW_METHODS; 
				}
		}
		
		/*=====================================================================
		  [Public] Property: Show Module Info 

		  Type:	Boolean  

		  Description:   Show the module information for any type found - this 
		                 is useful when trying to determine what DLL contains 
		                 what type.
		=====================================================================*/
		public bool ShowModuleInfo
		{
			get { return (showOptions & SHOW_MODULE_INFO) != 0; }
			set { 
					if ( value )
						showOptions |= SHOW_MODULE_INFO; 
					else
						showOptions &= ~SHOW_MODULE_INFO; 
				}
		}
		

		/*=====================================================================
		  [Public] Method: BuildDLLFileList

		  Description:   Given a directory this method appends the fullly 
		                 qualified path name of all DLLs in the directory to the
		                 list passed in.
		=====================================================================*/
		private void BuildDLLFileList(String directory, ArrayList list) 
		{
			try
			{

				if ( Directory.DirectoryExists(directory) )
				{
				
					Directory dir = new Directory(directory);
					File[] e = dir.GetFiles("*.dll");
			
					for(int i=0; i < e.Length; i++)
					{
						list.Add(String.Concat(directory, "\\", e[i].Name));
					} 
				}
				else
				{
					myVerboseWriter.WriteLine("Directory [{0}] does not exist!", directory);
				}
		
			}

			catch(Exception Ex)	 
			{

				//
				// Trap Exception: System.InvalidOperationException  This occurs when the file is an unmanaged dll
				//
				if (Ex.GetType() == typeof(System.InvalidOperationException))		
				{
				}					

			}			
		
		}

		
		/*=====================================================================
		  [Public] Method: Search 

		  Parameter(s)
				args Command Line arguments passed to the utility.

		  Description:   Sets the options specified by the user and then searches.
		=====================================================================*/
		public void Search(string[] args)
		{
			SetOptions( args );		
		}
		
		/*=====================================================================
		  [Public] Method: Search 

		  Parameter(s)
				         theSearchString The search string to use.

		  Description:   Searchs for the passed string in the type names of all
		                 classes in MSCORLIB (the standard System library) and
		                 all DLLs in the specified paths.
		=====================================================================*/
		public void Search(String theSearchString)
		{

			//
			// Search mscorlib
			//
			myVerboseWriter.WriteLine("Searching System Library (mscorlib)");				
			Search(theSearchString, "mscorlib.dll" );

			//
			// Search the current directory
			//
			myVerboseWriter.WriteLine("Searching the current directory...");
			ArrayList l = new ArrayList();
			BuildDLLFileList( "." , l );
			for ( int j = 0; j < l.Count; j++ )
			{
				Search(theSearchString, (String)l[j]);
			}
			
			//
			// Search the specified directories
			//
			Object[] dir = DirList.ToArray();
			for ( int i = 0; i < dir.Length; i++ )
			{
				myVerboseWriter.WriteLine("Searching directory  {0}...", dir[i]);
				
				l = new ArrayList();
				BuildDLLFileList( (String)dir[i], l );
				
				for ( int j = 0; j < l.Count; j++ )
				{
					Search(theSearchString, (String)l[j]);
				}
			}
		}
		
		/*=====================================================================
		  [Public] Method: Search 

		  Parameter(s)
				         theSearchString The search string to use.
				         theModule       The fully qualified path name of the Module to search
				                         for.

		  Description:   Loads the specified module and searchs through all the types
		                 defined in it for the type name specified.
		=====================================================================*/
		public void Search(String theSearchString, String theModule )
		{
			try 
			{
				//
				// Load the module - expect to fail if it is not a COM+ 2.0 
				// module.
				//

				Assembly a = Assembly.Load(theModule);
				Module[] m = a.GetModules();
				
				for ( int j = 0; j < m.Length; j++ )
				{

					myVerboseWriter.WriteLine("Searching Module {0}", theModule );
						
					if ( m!= null )
					{
						//
						// Get all the types from the module
						//
			    			Type[] types = m[j].GetTypes();
					
						for (int i=0; i < types.Length; i++)
						{
							Type curType = types[i];
						
							//
							// How has the user indicated they want to search. Note that 
							// even if the user has specified several ways to search we only 
							// respect one of them (from most specific to most general)
							//						
							if ( ExactMatchOnly )
							{
								if ( curType.FullName == theSearchString )
								{
									DumpType( curType );
								}												
							}
							else if ( MatchOnlyNamespace )
							{
								if ( curType.Namespace != null )
								{
									if ( curType.Namespace == theSearchString )
									{
										DumpType( curType );
									}						
								}							
							}
							else if ( WideSearch )
							{
								if ( curType.FullName.IndexOf(theSearchString) != -1 )
								{
									DumpType( curType );
								}																		
							}
							else
							{
								//
								// User has not specified a search criteria - so we have some 
								// defaults:
								//   (1) If the search string supplied matches a complete type
								//       name then assume they want to get all information about 
								//       the type. If they have actually set display options then 
								//       respect them.
								//   (2) If we are going to dump all information save the show 
								//       properties so they can be reset after this type (in case the 
								//       user is searching for multiple types).
								//
								if ( curType.FullName == theSearchString )
								{
									int oldOptions = showOptions;
								
									if ( showOptions == 0 )
									{
										ShowAll();
									}
									
									DumpType( curType );
									
									showOptions = oldOptions;
								}
								else if ( curType.Name.IndexOf(theSearchString) != -1 )
								{
									DumpType( curType );
								}
							}												
						}					
					}			
					
				}

			}
		
			catch(ReflectionTypeLoadException rcle)
			{

				Type[] loadedTypes     = rcle.Types;
				Exception[] exceptions = rcle.LoaderExceptions;
						
				int exceptionCount = 0;
				
				for ( int i =0; i < loadedTypes.Length; i++ )
				{
					if ( loadedTypes[i] == null )
					{
						myWriter.WriteLine("Unable to load a type because {0}", exceptions[exceptionCount] );
						exceptionCount++;		
					}
				}
			}
			catch(FileNotFoundException fnfe)
			{
				myVerboseWriter.WriteLine(fnfe.Message);
			}
			catch //(Exception Ex)
			{
				myVerboseWriter.WriteLine("Error reading {0}", theModule);
			}
		}
		
		/*=====================================================================
		  [Private] Method: GetTypeDescription

		  Parameter(s)
				         aType the type to use.
				         
		  Returns:  A short description of the type.
		=====================================================================*/
		private String GetTypeDescription(Type aType)
		{
			String str = null;

			if ( aType.IsClass )
			{
			  str = "class";
			}                                 
			
			if ( aType.IsInterface )
			{
			  str = "interface";
			}                                 
			
			if ( aType.IsValueType )
			{
			  str = "struct";
			} 
		       
			if ( aType.IsArray )
			{
			  str = "array";
			}                                 
			
			return str;
		}
		
		/*=====================================================================
		  [Private] Method: DumpType 

		  Parameter(s)
				aType the type to dump.

		  Description:   Dumps information about the specified type.
		=====================================================================*/
		private void DumpType(Type aType)
		{  
			Type baseType = aType.BaseType;

			myWriter.WriteLine("{0,-10} {1}", GetTypeDescription(aType), aType );
			
			if ( ShowModuleInfo )
			{
				myWriter.WriteLine("{0,-10} {1}", "Module:", aType.Module.FullyQualifiedName );
			}
						
			DumpInterfaces( aType );
			DumpFields( aType );
			DumpProperties( aType );
			DumpEvents( aType );			
			DumpMethods( aType );
			
			if ( RecurseTypes )
			{
				myWriter.WriteLine();
			}	                                        
				          
			//
			// If recursing then pop the indent on the writers so we 
			// can easily see the nesting for the base type information.
			//	                             	    
		    if ( RecurseTypes && baseType != null )
			{
				myWriter.PushIndent();
				myVerboseWriter.PushIndent();
					
				DumpType( baseType );
					
				myWriter.PopIndent();
				myVerboseWriter.PopIndent();
			}			
		}
		  
		/*=====================================================================
		  [Private] Method: DumpInterfaces

		  Parameter(s)
				aType the type to dump.

		  Description:   Dumps the interfaces implemented by the specified type.
		=====================================================================*/
		private void DumpInterfaces(Type aType)
		{
			if ( !ShowInterfaces )
				return;
				
			Type[] info = aType.GetInterfaces();
			
			if ( info.Length != 0 )
			{					    					    
				myWriter.WriteLine("{0} {1}", "# Interfaces:", 
				                          info.Length ); 
				  
				for ( int i=0; i < info.Length; i++ )
				{
				
				  myWriter.WriteLine("  interface {0}", info[i].FullName);
				  
				  //
				  // Only show method information only if requested
				  //
			      if ( ShowMethods )
				  {
					myWriter.PushIndent();
					DumpType( info[i] );
					myWriter.PopIndent();			  
				  }
				}
			}
		}

		/*=====================================================================
		  [Private] Method: DumpProperties

		  Parameter(s)
				aType the type to dump.

		  Description:   Dumps the public properties directly contained in the 
		                 specified type.
		=====================================================================*/
		private void DumpProperties(Type aType)
		{
			if ( !ShowProperties )
				return;
				
			PropertyInfo[] pInfo = aType.GetProperties();
			myWriter.WriteLine("Properties");
			
			bool found = false;
			
			if ( pInfo.Length != 0 )
			{						                                     
				PropertyInfo curInfo = null;
												  
				for ( int i=0; i < pInfo.Length; i++ )
				{
				  curInfo = pInfo[i];
				
				  //
				  // Only display properties declared in this type.
				  //		    
    			  if ( curInfo.DeclaringType != aType )
    			  {
    				continue;
    			  }
    			  
    			  found = true;
    			  
				  String flags = null;
						    
				  if ( curInfo.CanRead && curInfo.CanWrite )
				  {
					flags = "get; set;";
				  }				  		    
				  else if ( curInfo.CanRead )
				  {
				    flags = "get";
				  }
				  else if ( curInfo.CanWrite )        
				  {
				    flags = "set";
				  }
				  
				  myWriter.WriteLine("  {0,-10} '{1}' ", curInfo, 
												  flags);
				}
			}
			
			if ( !found )
			{
				myWriter.WriteLine("  (none)");
			}					  
		}
	  
		/*=====================================================================
		  [Private] Method: DumpEvents

		  Parameter(s)
				aType the type to dump.

		  Description:   Dumps the public events directly contained in the 
		                 specified type.
		=====================================================================*/
		private void DumpEvents(Type aType)
		{
			if ( !ShowEvents )
				return;
				
			EventInfo[] eInfo = aType.GetEvents( );
				
			myWriter.WriteLine("Events:");
			bool found = false;
			
			if ( eInfo.Length != 0 )
			{					                                     
				for ( int i=0; i < eInfo.Length; i++ )
				{        
  			      //
				  // Only display events declared in this type.
				  //		    
				  if ( eInfo[i].DeclaringType == aType )
				  {
					found = true;
				    myWriter.WriteLine("  {0}", eInfo[i]);
				  }
				}
			}
			
			if ( !found )
			{
				myWriter.WriteLine("  (none)");
			}					  

		}

		/*=====================================================================
		  [Private] Method: DumpFields

		  Parameter(s)
				aType the type to dump.

		  Description:   Dumps the public fields directly contained in the 
		                 specified type.
		=====================================================================*/
		private void DumpFields(Type aType)
		{
			if ( !ShowFields )
				return;

			FieldInfo[] info = aType.GetFields( );
				
			myWriter.WriteLine("Fields:");
			
			bool found = false;
			
			if ( info.Length != 0 )
			{
				for ( int i=0; i < info.Length; i++ )
				{        
				  //
				  // Only display fields declared in this type.
				  //		    
				  if (info[i].DeclaringType == aType )
				  {
				    myWriter.WriteLine("  {0} {1}", info[i], info[i].Name );
				    found = true;
				  }
				}
			}		    
			
			if ( !found )
			{
				myWriter.WriteLine("  (none)");
			}					  			
		}
	  
		/*=====================================================================
		  [Private] Method: DumpMethods

		  Parameter(s)
				aType the type to dump.

		  Description:   Dumps the public methods directly contained in the 
		                 specified type. Note "special name" methods are not 
		                 displayed.
		=====================================================================*/
		private void DumpMethods(Type aType)
		{
			if ( !ShowMethods )
				return;

			MethodInfo[] mInfo = aType.GetMethods();
			
			myWriter.WriteLine("Methods"); 
			
			bool found = false;			
			
			if ( mInfo.Length != 0 )
			{
				for ( int i=0; i < mInfo.Length; i++ )
				{
				  //
				  // Only display methods declared in this type. Also 
				  // filter out any methods with special names - these
				  // cannot be generally called by the user (i.e their 
				  // functionality is usually exposed in other ways e.g
				  // property get/set methods are exposed as properties.
				  //		    
			      if ( mInfo[i].DeclaringType == aType && 
					   !mInfo[i].IsSpecialName 	
			         )
			      {		
					found = true;
						 
					StringBuilder modifiers = new StringBuilder();
					
					if ( mInfo[i].IsStatic )   { modifiers.Append("static "); }     
					if ( mInfo[i].IsPublic )   { modifiers.Append("public "); }     
					if ( mInfo[i].IsFamily )   { modifiers.Append("protected "); }     
					if ( mInfo[i].IsAssembly ) { modifiers.Append("internal "); }     
					if ( mInfo[i].IsPrivate )  { modifiers.Append("private "); }     
					
				    myWriter.WriteLine("  {0} {1}", modifiers ,mInfo[i]);
				  }
				}					  
			}				    
			
			if ( !found )
			{
				myWriter.WriteLine("  (none)");
			}
		}
	
		/*=====================================================================
		  [Private] Method: ShowAll

		  Description:   Sets all display options on.
		=====================================================================*/
		private void ShowAll()
		{
			ShowInterfaces = true;
			ShowMethods    = true;
			ShowFields     = true;
			ShowProperties = true;
			ShowEvents     = true;
			ShowModuleInfo = true;
		}
		
		/*=====================================================================
		  [Private] Method: PrintUsage

		  Parameter(s)
				aType the type to dump.

		  Description:   Prints out usage information to the user.
		=====================================================================*/
		private void PrintUsage()
		{
			myWriter.WriteLine();
			
			myWriter.WriteLine("FindClass [SystemOptions] [MatchOptions] [ShowOptions] [MiscOptions] Name");
			myWriter.WriteLine("");
			myWriter.WriteLine("  where SystemOptions");
			myWriter.WriteLine("	d:[Directory] - Additional directory to search");

			myWriter.WriteLine("  where MatchOptions");
			myWriter.WriteLine("    x - Name = Exact Type Name (including namespace)");
			myWriter.WriteLine("    n - Name = Namespace ");
			myWriter.WriteLine("    w - Match the name anywhere in the namespace");

			myWriter.WriteLine("  where ShowOptions");						
			myWriter.WriteLine("    i - Show Interfaces");
			myWriter.WriteLine("    f - Show Fields");
			myWriter.WriteLine("    p - Show Properties");
			myWriter.WriteLine("    e - Show Events");
			myWriter.WriteLine("    m - Show Methods");
			myWriter.WriteLine("    a - Show All Info");
			myWriter.WriteLine("    l - Show Module Information");
			
			myWriter.WriteLine("  where MiscOptions");									
			myWriter.WriteLine("    v = Verbose");		
			myWriter.WriteLine("    r = For every type find display base type information");		
			myWriter.WriteLine("    ? = Prints Usage information");		
			
			myWriter.WriteLine();
			
			myWriter.WriteLine("Examples");
			myWriter.WriteLine();
			
			myWriter.WriteLine(" FindType String");
			myWriter.WriteLine("   Finds all types that have 'String' as part of their type name\n");
			myWriter.WriteLine(" FindType -r String");
			myWriter.WriteLine("   Finds all types that have 'String' as part of their name");
			myWriter.WriteLine("   and prints out all base classes\n");
			myWriter.WriteLine(" FindType -n System.Reflection");
			myWriter.WriteLine("   Displays all types in the 'System.Reflection' namespace\n");
			myWriter.WriteLine(" FindType -xipm System.String");
			myWriter.WriteLine("   Displays the interfaces, properties and methods on the 'System.String' type\n");
			myWriter.WriteLine(" FindType -d:C:\\ -d:\"C:\\Program Files\\COM20SDK\\Lib\" String");
			myWriter.WriteLine("   Searches DLLs C:\\ and C:\\Program Files\\COM20SDK\\Lib");
			myWriter.WriteLine("   as well as in the current directory \n");
			
			myWriter.WriteLine();
		}


		/*=====================================================================
		  [Private] Method: SetOptions

		  Description:   Processes all of the options specified in the arguments
		=====================================================================*/
		private void SetOptions(string[] args)
		{
			if ( args.Length == 0 )
			{
				PrintUsage();
				return;
			}
			
			myVerboseWriter.Print = false;

			char[] backslash = new char[1];
			backslash[0] = '\\';
			
			for( int i = 0; i < args.Length; i++ )
			{
				string curArg = args[i];
				
				if ( curArg[0] == '-' || curArg[0] == '/'  )
				{
					if (Char.ToUpper(curArg[1]) == 'D')
					{
						if (curArg.Length > 2 && curArg[2] == ':')
						{
							String dir = curArg.Substring(3).TrimEnd(backslash).ToUpper();

							if (dir != "")
								DirList.Add(dir);
							else
								myWriter.WriteLine("Directory not specified");
						}
						else
							myWriter.WriteLine("Directory not specified");
					}
					else
					{
						for ( int j = 1; j < curArg.Length; j++ )
						{
							switch( Char.ToUpper(curArg[j]) )
							{
								case 'X':
									ExactMatchOnly = true;
								break;
								case 'R':
									RecurseTypes = true;
								break;
								case 'V':
									myVerboseWriter.Print = true;
								break;
		    					case 'N':
									MatchOnlyNamespace = true;
								break;
		    					case 'W':
									WideSearch = true;
								break;
		    					case 'I':
									ShowInterfaces = true;
								break;
		    					case 'M':
									ShowMethods = true;
								break;
		    					case 'F':
									ShowFields = true;
								break;							
		    					case 'P':
									ShowProperties = true;
								break;							
		    					case 'E':
									ShowEvents = true;
								break;							
		    					case 'L':
									ShowModuleInfo = true;
								break;							
		    					case 'A':
		    						ShowAll();
								break;							
		    					case '?':
		    						PrintUsage();
								break;			
								case 'D':
									myWriter.WriteLine("Directory not specified");
								break;
								default:
								{
									myWriter.WriteLine("Invald Option[{0}]", curArg[j] );
								}
								break;
							}
						}
					}				
				}
				else
				{
					myClassList.Add( curArg );
				}
			}	
			
			if ( myClassList.Count != 0 )
			{
				for ( int i = 0; i < myClassList.Count; i++ )
				{
					Search( (string)myClassList[i] );
				}
			}
		}
					
		/*=====================================================================
		  [Public] Method: Main

		  Parameter(s)
				args The arguments passed to this utility

		  Description:   Instantiates the FindType class and processes the user 
						 command line options.
		=====================================================================*/
	    public static void Main(String[] args) 
		{	
			FindType f = new FindType();
			f.Search( args );
		} 
	}
}
