/*=====================================================================
  File:      Perms.cs

  Summary:   Demonstrates how use code access security.

---------------------------------------------------------------------
  This file is part of the Microsoft COM+ 2.0 SDK Code Samples.

  Copyright (C) 2000 Microsoft Corporation.  All rights reserved.

This source code is intended only as a supplement to Microsoft
Development Tools and/or on-line documentation.  See these other
materials for detailed information regarding Microsoft code samples.

THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
PARTICULAR PURPOSE.
=====================================================================*/


// Add the classes in the following namespaces to our namespace
using System;
using System.IO;
using System.Security.Permissions;
using System.Security;


///////////////////////////////////////////////////////////////////////////////


// This class represents the application itself
class Application {
    public static void Main() {

        // Try to access resources using the permissions currently available.
        AttemptAccess("Default permissions");



        // Create a permission set that allows read access to the TEMP 
        // environment variable and read, write, and append access to SomeFile
        PermissionSet ps = new PermissionSet(PermissionState.None);
        ps.AddPermission(
            new EnvironmentPermission(EnvironmentPermissionAccess.Read, "TEMP"));
        ps.AddPermission(
            new FileIOPermission(FileIOPermissionAccess.Read | 
                FileIOPermissionAccess.Write | FileIOPermissionAccess.Append, 
                "SomeFile"));


        // Make the permissions indicate all that we're allowed to do.
        ps.Assert();

        // Try to access resources using the permissions we've just asserted.
        AttemptAccess("Assert permissions");

        // Remove this stack frame's Assert so we're back to default permissions.
        CodeAccessPermission.RevertAssert();


        // Deny access to the resources we specify
        ps.Deny();

        // Try to access resources using the permissions we've just denied.
        AttemptAccess("Deny permissions");

        // Remove this stack frame's Deny so we're back to default permissions.
        CodeAccessPermission.RevertDeny();



        // Make the permissions indicate the only things that we're allowed to do.
        ps.PermitOnly();

        // Try to access resources using only the permissions we've just permitted.
        AttemptAccess("PermitOnly permissions");

        // Remove this stack frame's PermitOnly so we're back to default permissions.
        CodeAccessPermission.RevertPermitOnly();



        // Remove the FileIOPermissions from the permission set
        ps.RemovePermission(typeof(FileIOPermission));


        // Try to access resources using only the Environment permissions.
        ps.PermitOnly();
        AttemptAccess("PermitOnly without FileIOPermission permissions");
        CodeAccessPermission.RevertPermitOnly();

        // Remove the EnvironmentPermission from the permission set
        ps.RemovePermission(typeof(EnvironmentPermission));


        // Try to access resources using no permissions.
        ps.PermitOnly();
        AttemptAccess("PermitOnly without any permissions");
        CodeAccessPermission.RevertPermitOnly();


        // DEMO: Show how to use Demand/Assert to improve performance
        CopyFile(".\\Permissions.exe", ".\\Permissions.copy.exe");
    }


    static public void AttemptAccess(String s) {
        FileStream fs = null;
        String ev = null;

        // Try to access a file
        try {
            fs = new FileStream("SomeFile", FileMode.OpenOrCreate); 
        }
        catch (Exception) { 
        }

        // Try to read an environment variable
        try {
            ev = Environment.GetEnvironmentVariable("TEMP");
        }
        catch (Exception) { 
        }

        // Display what we sucessfully did.
        Console.WriteLine(s + " test: " 
            + ((fs != null) ? "FileOpen" : "FileNOTOpened") + ", " 
            + ((ev != null) ? "EVRead" : "EVNOTRead"));

        // If we opened the file, close it & delete it
        if (fs != null) { 
            fs.Close(); 
            File.Delete("SomeFile");
        }
    }


    // BUGBUG: Verify this code w/LorenK
    // BUGBUG: Does the actual copy portion have to go into its own method?
    // BUGBUG: Add comments explaining stack frame performance
    public static void CopyFile(String srcPath, String dstPath) {

        // Create a file permission set indicating all of this method's intentions.
        FileIOPermission fp = new FileIOPermission(FileIOPermissionAccess.Read, srcPath);
        fp.AddPathList(FileIOPermissionAccess.Write | FileIOPermissionAccess.Append, dstPath);

        // Verify that we can be granted all the permissions we'll need.
        fp.Demand();

        // Assert the desired permissions here.
        fp.Assert();

        // For the remainder of this method, demands for source file read access
        // and demands for destination file write/append access will be granted
        // immediately; walking the remainder of the stack will not be necessary.

        File srcFile = new File(srcPath);
        File dstFile = new File(dstPath);
        Stream src = srcFile.Open(FileMode.Open, FileAccess.Read, FileShare.Read);
        Stream dst = dstFile.Open(FileMode.Create, FileAccess.Write, FileShare.None);
        if (srcFile.Length > Int32.MaxValue)
            throw new Exception("CopyFile requires that the source file be less than 2GB.");
        Byte[] buffer = new Byte[(Int32) srcFile.Length];
        src.Read(buffer, 0, (Int32) srcFile.Length);
        dst.Write(buffer, 0, (Int32) srcFile.Length);
        //dstFile.CreationTime   = srcFile.CreationTime;
        //dstFile.LastAccessTime = srcFile.LastAccessTime;
        //dstFile.LastWriteTime  = srcFile.LastWriteTime;
        src.Close();
        dst.Close();
    }
}


///////////////////////////////// End of File /////////////////////////////////
